;; Generated package description from sql-indent.el  -*- no-byte-compile: t -*-
(define-package "sql-indent" "1.6.0.20220424.94230" "Support for indenting code in SQL files." '((cl-lib "0.5")) :commit "5a3ce9e52cc3fb55ec02f3a38f43d687a9e3672d" :authors '(("Alex Harsanyi" . "AlexHarsanyi@gmail.com")) :maintainer '("Alex Harsanyi" . "AlexHarsanyi@gmail.com") :keywords '("languages" "sql") :url "https://github.com/alex-hhh/emacs-sql-indent")
